#!/usr/bin/python

"""
  register automata model
"""

# Author: Fides Aarts & Harco Kuppens

import cgi
import xml.etree.cElementTree as ElementTree
import re
import tempfile, os
import pyuppaal
import model_parser

class Register:
    def __init__(self, alphabet=None, constants=None, globalVars=None, locations=None, transitions=None, initLocation=None):
        self.alphabet = alphabet
        self.constants = constants or []
        self.globalVars = globalVars or []
        self.locations = locations or []
        self.transitions = transitions or []
        self.initLocation = initLocation or None

    def to_file(self,filename):
         f=open(filename,'w')
         f.write(self.to_xml())
         f.close()

    def to_dotfile(self,filename):
         f=open(filename,'w')
         f.write(self.to_dot())
         f.close()

    def to_uppaalfile(self,filename):
         f=open(filename,'w')
         f.write(self.to_uppaal().to_xml())
         f.close()         


    def to_xml(self):

        xml_constants="\n".join([c.to_xml() for c in self.constants])
        xml_globals="\n".join([g.to_xml() for g in self.globalVars])
        xml_locations="\n".join([l.to_xml() for l in self.locations])
        xml_transitions="\n".join([t.to_xml() for t in self.transitions])
        xml="""<?xml version="1.0" encoding="UTF-8" ?>
<register-automaton>
   <alphabet>
   %s
   </alphabet>
   <constants>
%s
   </constants>
   <globals>
%s
   </globals>
   <locations>
%s
   </locations>
   <transitions>
%s
   </transitions>
</register-automaton>
""" % (self.alphabet.to_xml(),xml_constants,xml_globals,xml_locations,xml_transitions)  
        return xml
    
    def to_dot(self):
        locations="\n".join([l.to_dot() for l in self.locations])
        transitions="\n".join([t.to_dot() for t in self.transitions])
        dotstr="digraph g {\nnode [shape=oval];\n%s%s\n}\n" % (locations,transitions)  
        return dotstr

    def to_uppaal(self):

        system_str='Process = Template();system Process;'
        decl_str="\n"
        
#        #constants = 0,1;
#        constlist=[str(c.value)  for c in self.constants]
#        if constlist:
#            decl_str += "constants = " + ",".join(constlist) + ";\n"
            
        for constVar in self.constants:
            decl_str += constVar.getDeclStr() +"\n";    

        for globalVar in self.globalVars:
            decl_str += globalVar.getDeclStr() +"\n";    
        
        for symb in self.alphabet.inputs + self.alphabet.outputs :
            decl_str += symb.getDeclStr()    
        
        model= pyuppaal.NTA(declaration=decl_str,system=system_str)

        # TODO fill these
        
        initloc=pyuppaal.Location(id=self.initLocation.getId()) 
        
        locs=[]
        id2location={}
        for loc in self.locations:
           location=pyuppaal.Location(id=loc.getId()) 
           locs.append( location )
           id2location[loc.getId()]=location
           
        trans=[]
        for tra in self.transitions:
             methodCall=tra.getMethodCallString()
             upp_trans=pyuppaal.Transition( 
                id2location[tra.source.getId()], 
                id2location[tra.target.getId()], 
                select='', 
                guard=tra.getGuard(), 
                synchronisation=methodCall,
                assignment=tra.getAction()
             )  
             trans += [upp_trans] 

        template=pyuppaal.Template(name='Template', locations=locs, initlocation=initloc, transitions=trans)
        model.add_template(template)

        return model 

class Alphabet:
    def __init__(self, inputs=[], outputs=[] ):
        self.inputs = inputs
        self.outputs = outputs

    def to_xml(self):
        xml_inputs="\n".join([i.to_xml() for i in self.inputs])
        xml_outputs="\n".join([o.to_xml() for o in self.outputs])
        xml="""   <inputs>
%s
      </inputs>
      <outputs>
%s
      </outputs>""" % (xml_inputs,xml_outputs)      
        return xml

class Symbol:
    def __init__(self, name, params=[] ):
        self.name = name
        self.params = params
        
    def getDeclStr(self):
        params_as_declstrings=[]
        param_str=""
        for i in range(0,len(self.params)):
            params_as_declstrings += [ "%s p%s" %  ( self.params[i], str(i) ) ]
        param_str= ", ".join(params_as_declstrings)
        
        return "void %s(%s) {\n}\n" % (self.name,param_str)
    
    def to_xml(self):
        #print self.params
        xml_params=""
        if len(self.params) > 0:
            for i in range(0,len(self.params)):
                xml_params+= """\n            <param type="%s" name="p%s"/>""" % ( cgi.escape(self.params[i]), cgi.escape(str(i)) )
            xml="""         <symbol name="%s"> %s \n         </symbol>""" % (cgi.escape(self.name),xml_params)
        else:
            xml="""         <symbol name="%s"/>""" % (cgi.escape(self.name))
        return xml       
    
class Constant:
    def __init__(self, name, value, vartype="int" ):
        self.name = name
        self.value = value
        self.vartype = vartype

    def getDeclStr(self):
        return "const %s %s = %s;" % (self.vartype,self.name,self.value)

    def to_xml(self):
        return """      <constant type="%s" name="%s">%s</constant>""" % (cgi.escape(self.vartype),cgi.escape(self.name),cgi.escape(str(self.value)))

class GlobalVar:
    def __init__(self, name, value, vartype="int" ):
        self.name = name
        self.value = value
        self.vartype = vartype

    def getDeclStr(self):
        return "%s %s = %s;" % (self.vartype,self.name,self.value)
    
    def to_xml(self):
        return """      <variable type="%s" name="%s">%s</variable>""" % (cgi.escape(self.vartype),cgi.escape(self.name),cgi.escape(str(self.value)))

class Location:
    def __init__(self, name, initial=False ):
        self.name = name
        self.initial = initial

    def getId(self):
        return self.name
    
    def to_xml(self):
        if self.initial == True:
            return """      <location name="%s" initial="true"/> """ % (cgi.escape(self.name))
        else:
            return """      <location name="%s"/> """ % (cgi.escape(self.name))
        
    def to_dot(self):
        if self.initial == True:
            return "  %s [peripheries=2,color=\"red\"];" % (self.name)
        else:
            return "  %s;" % (self.name)

class Transition:
    def __init__(self, source, target, method, param_names = [], guard='', action_statements=[]):
        self.source = source
        self.target = target 
        self.method = method
        self.param_names = param_names or []
        self.guard = guard or ""
        self.action_statements = action_statements or [] # assignments only for register
    def getMethodCallString(self):
        param_str=",".join(self.param_names)
        return "%s(%s)" % (self.method, param_str)
    
    def getAction(self):
        if self.action_statements == []:
          return ""
        else:
          return ";".join(self.action_statements) + ";"
    
    def getGuard(self):
        return self.guard

    def to_xml(self):
        if self.guard:
            xml_guard="\n         <guard>%s</guard>" % (cgi.escape(self.guard))
        else:
            xml_guard=""            
        if self.action_statements:
            xml_assignTo=""
            for action_statement in self.action_statements:
                split_action_statement = action_statement.split("=",1);
                xml_assignToVariable = split_action_statement[0].strip()
                #print "xml_assignTo: " + xml_assignToVariable
                xml_assignToValue = split_action_statement[1]
                #print "xml_assignToValue: " + xml_assignToValue
                xml_assignTo += """\n            <assign to="%s">%s</assign>""" % (cgi.escape(xml_assignToVariable),cgi.escape(xml_assignToValue))
            xml_assignments="\n         <assignments>%s\n         </assignments>" % (xml_assignTo)
        else:   
            xml_assignments=""
        
        xml_params=",".join(self.param_names)
        if xml_params:
            xml_params='params="%s"' % (xml_params)
        else:
            xml_params=""
                   
        if xml_guard or xml_assignments:
            xml="""      <transition from="%s" to="%s" symbol="%s" %s>%s%s\n      </transition>""" % (self.source.getId(),self.target.getId(),cgi.escape(self.method),xml_params,xml_guard,xml_assignments)
        else:  
            xml="""      <transition from="%s" to="%s" symbol="%s" %s/>""" % (self.source.getId(),self.target.getId(),cgi.escape(self.method),xml_params)
        
        return xml  

    def to_dot(self):
        xml_guard=""
        xml_assignments=""
        if self.guard:
            xml_guard=self.guard
        if self.action_statements:
            xml_assignments=";".join(self.action_statements)
        xml_params=",".join(self.param_names)
        if not xml_params:
            xml_params=""
        label="%s%s%s" %  (self.method,xml_guard,xml_assignments)
        xml="  %s -> %s [label=\"%s\"];" % (self.source.getId(),self.target.getId(),label)                  
        return xml  


    #def to_xml(self):
    #    return """
    #<transition>
    #  <source ref="%s"/>
    #  <target ref="%s"/>
    #</transition>""" % (self.source.id, self.target.id)


#===============================================================================
#                       pretty printer xml tree node         
#===============================================================================

def indent(elem, level=0):
    i = "\n" + level*"  "
    if len(elem):
        if not elem.text or not elem.text.strip():
            elem.text = i + "  "
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
        for elem in elem:
            indent(elem, level+1)
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
    else:
        if level and (not elem.tail or not elem.tail.strip()):
            elem.tail = i
            
def getPrettyPrintedString(treeNode):
    indent(treeNode)
    return  xml.etree.ElementTree.tostring(treeNode)



def from_xml(filename):
    xmlrootelem = ElementTree.ElementTree(file=filename).getroot()
    # see doc at : http://docs.python.org/2/library/xml.etree.elementtree.html

    #print getPrettyPrintedString(xmlrootnode)


    initLocation=None
    locations = []
    id2location={}
    for locationxml in xmlrootelem.iter("location"):
        name = locationxml.attrib["name"]
        
        if locationxml.attrib.has_key("initial") and locationxml.attrib["initial"] == "true":
           location = Location(name,initial=True)
           initLocation=location
        else: 
           location = Location(name,initial=False)
        #print location.to_xml()
        locations += [location]
        id2location[location.getId()]=location
        
    transitions = []
    for transitionxml in xmlrootelem.iter("transition"):
        tr_from=transitionxml.attrib["from"]
        tr_to=transitionxml.attrib["to"]
        tr_symbol=transitionxml.attrib["symbol"]
        tr_param_names=[]
        if transitionxml.attrib.has_key("params"):  
            paramsstr=transitionxml.attrib["params"]           
            tr_param_names=[ param.strip() for param in paramsstr.split(',')  ]
        guardElem=transitionxml.find("guard")
        tr_guard=""
        if guardElem is not None:
             tr_guard=guardElem.text
        assigns = []
        for assignxml in transitionxml.iter("assign"): 
            assign_to=assignxml.attrib["to"] 
            assign_value=assignxml.text 
            assigns  += [ assign_to + "=" + assign_value ]      
        
        transition = Transition(
            id2location[tr_from],
            id2location[tr_to],
            tr_symbol, 
            tr_param_names,
            tr_guard,
            assigns 
        )
        transitions += [transition]
    
    constants=[]    
    constantsElem=xmlrootelem.find("constants")
    if constantsElem is not None:
        for constantxml in constantsElem:
            c_type=constantxml.attrib["type"]
            c_name=constantxml.attrib["name"]
            c_value=constantxml.text
            constants += [Constant(c_name,c_value,c_type)] 
        
        
    globalVars=[]    
    globalsElem=xmlrootelem.find("globals")
    if globalsElem is not None:
        for globalxml in globalsElem:
            c_type=globalxml.attrib["type"]
            c_name=globalxml.attrib["name"]
            c_value=globalxml.text
            globalVars += [GlobalVar(c_name,c_value,c_type)]         
      

    
    def getSymbols(xmlElem):
        symbols=[]
        if not xmlElem:
            return []
        for symbol in xmlElem:
            params=[]
            for param in symbol:
                params += [ param.attrib["type"] ]
                # note parameter names not important for symbol type declaration in alphabet
            symbols += [ Symbol(symbol.attrib["name"],params)]     
        return symbols
    
    elemAlphabet=xmlrootelem.find("alphabet")        
    inputs=getSymbols(elemAlphabet.find("inputs"))
    outputs=getSymbols(elemAlphabet.find("outputs"))   
    alphabet=Alphabet(inputs, outputs )
    
  
    
    
    reg = Register(alphabet, constants, globalVars, locations, transitions,initLocation)
    return reg
    #def __init__(self, alphabet=None, constants=None, globalVars=None, locations=None, transitions=None):


def from_uppaal_modelfile(uppaal_modelfile):
    uppaal_model=pyuppaal.from_xml(uppaal_modelfile)
    return from_uppaal_model(uppaal_model)

def from_uppaal_model(uppaal_model):

    # parse data from declarations
    data=model_parser.parse_declarations(uppaal_model.declaration)
    
    # get data from uppaal model
    template=uppaal_model.templates[0]
    uppaal_inputs=[input for input in data['inputInterfaces'].keys()]
    uppaal_outputs=[output for output in data['outputInterfaces'].keys()]
    #uppaal_constantValues=[constantValue for constantValue in data['constants']]
    uppaal_constants=[constant for constant in data['constants'].keys()]
    uppaal_statevars=[statevar for statevar in data['statevars'].keys()]
    uppaal_locations=[loc.id for loc in template.locations]
    uppaal_initlocation=template.initlocation.id
    uppaal_transitions=[transition for transition in template.transitions]

    # TODO
    inputs=[]
    for input in uppaal_inputs:
        params=[]
        #print input
        for param in data['inputInterfaces'][input]:
            if param.lower() == "integer":
                param = "int"
            # else take value set  
            params += [param]
            #print param
        inputs += [Symbol(input,params)]
        
    outputs=[]
    for output in uppaal_outputs:
        params=[]
        #print output
        for param in data['outputInterfaces'][output]:
            if param.lower() == "integer":
                param = "int"
            # else take value set  
            params += [param]
            #print param
        outputs += [Symbol(output,params)]
    
#    constants=[]  
#    i=0  
#    for constantValue in uppaal_constantValues:
#        #print constantValue
#        constantName = "constant"+str(i)
#        constants += [Constant(constantName,constantValue)]
#        i+=1
     
    constants=[]     
    for constantName in uppaal_constants:
        constantValue = data['constants'][constantName][1]
        constants += [Constant(constantName,constantValue)]

    globalVars=[]     
    for statevarName in uppaal_statevars:
        #print statevarName
        statevarValue = data['statevars'][statevarName][1]
        globalVars += [GlobalVar(statevarName,statevarValue)]
        
    locations=[]
    initLocation=None
    id2location={} 
    for loc in uppaal_locations:
        #print loc
        #print  uppaal_initlocation
       
        if loc == uppaal_initlocation: 
            location=Location(loc, True)  
            initLocation=location                          
        else:
            location=Location(loc, False)   
        locations += [location]    
        id2location[loc] = location    
        
    transitions=[]
    for transition in uppaal_transitions:
         #print transition
         source=transition.source.id
         #print "source: " + source
         #io_event=transition.io_event
         io_event=transition.synchronisation.get_value().strip()
         [io_event_name,io_type,param_names]=model_parser.parse_io_event(io_event)
         #print "io_event_name: " + io_event_name
         guard=transition.guard.get_value()
         #guard=model_parser.relabel_basic_params(guard,param_names)
         #print "guard: " + guard
         action=transition.assignment.get_value()
         action_statements = model_parser.parse_action(action)
         #action_statements=[ model_parser.relabel_basic_params(stmt,param_names) for stmt in  action_statements]
         #print "action: " + action
         #for action_statement in action_statements:
         #   print "action_statement: " + action_statement
         target=transition.target.id
         #print "target: " + target
         transitions += [Transition(id2location[source], id2location[target], io_event_name, param_names,  guard, action_statements)]
    
    #alphabet=Alphabet(inputs)
    alphabet=Alphabet(inputs, outputs)
    #globalVars=None
    #globalVars=[ GlobalVar(statevar) for statevar in uppaal_statevars]
    #locations=[ Location(loc) for loc in uppaal_locations]
    #initlocation= Location(uppaal_initlocation)
    #transitions=None
    reg = Register(alphabet, constants, globalVars, locations, transitions,initLocation)
    return reg


def from_automatonType_xml(filename):
    xmlrootelem = ElementTree.ElementTree(file=filename).getroot()
    # see doc at : http://docs.python.org/2/library/xml.etree.elementtree.html

    #print getPrettyPrintedString(xmlrootnode)

#  <automatonType helper="de.ls5.jlearn.util.xml.SimpleXmlSerializationHelper" version="1.0">
#     <state id="2"/>

    initLocation=None
    locations = []
    id2location={}
    for locationxml in xmlrootelem.iter("state"):
        name = locationxml.attrib["id"]
         
        if locationxml.attrib.has_key("initial") and locationxml.attrib["initial"] == "true":
           location = Location(name,initial=True)
           initLocation=location
        else: 
           location = Location(name,initial=False)
        #print location.to_xml()
        locations += [location]
        id2location[location.getId()]=location
        
#     <transition dest="5" source="0">
#         <input>B</input>
#         <output>W</output>
#     </transition>        
        
    transitions = []
    for transitionxml in xmlrootelem.iter("transition"):
        tr_from=transitionxml.attrib["source"]
        tr_to=transitionxml.attrib["dest"]
        
        outputelem=transitionxml.find("output")       
        if outputelem is None: 
            output="Q"  # quiescence
        else:
            output=outputelem.text 
           
        tr_symbol=transitionxml.find("input").text + "/" + output
        tr_param_names=[]
        tr_guard=""
        assigns = []
         
        transition = Transition(
            id2location[tr_from],
            id2location[tr_to],
            tr_symbol, 
            tr_param_names,
            tr_guard,
            assigns 
        )
        transitions += [transition]
     
    constants=[]    
    globalVars=[]    

#    <alphabet>
#        <input>A</input>
#        <input>B</input>
#        <input>C</input>
#        <input>D</input>
#        <input>E</input>
#        <input>F</input>
#    </alphabet>
    def getSymbols(xmlElem):
        symbols=[]
        if not xmlElem:
            return []
        for symbol in xmlElem:
            params=[]  # symbols without params
            symbols += [ Symbol(symbol.text,params)]     
        return symbols
     
    elemAlphabet=xmlrootelem.find("alphabet")        
    inputs=getSymbols(elemAlphabet)
    outputs=[]  # not defined in source
    alphabet=Alphabet(inputs,outputs)
     
    reg = Register(alphabet, constants, globalVars, locations, transitions,initLocation)
    return reg


if __name__ == "__main__":
    import sys
    inputfile="testinput/sicco_lts_mealy_competition.xml"
    dot_outputfile="out.dot"
    pdf_outputfile="out.pdf"
    
    regmodel=from_automatonType_xml(inputfile)
    #sys.stdout.write(regmodel.to_xml())
    #sys.stdout.write(regmodel.to_dot())
    regmodel.to_dotfile(dot_outputfile)

#    import pydot
#    graph=pydot.graph_from_dot_file(dot_outputfile)
#    graph.write_pdf(pdf_outputfile)
    
    
    
